#' \code{plotSignal} plots a matrix of signals
#' 
#' @param Signal a data.frame with columns "index" (numeric, the block numbering), "cab" (factor, indicating the capture antibody names), "signal" (numeric, the signal values), "lysate" (numeric, the fraction of lysate on the block), "error" (numeric, the standard error of the signal). If the data.frame contains "sigma", it is assumed that sigma represents a better estimate of the standard error based on an error model (see \code{getErrorModel})
#' @param cab character, the names of the capture antibodies to be plotted,
#' all if nothing is specified
#' @param index numeric, an index vector of the blocks to be plotted.
#' Defaults to all blocks.
#' @return a ggplot2 object
#' @export
plotSignal <- function(Signal, cabs = unique(Signal$cab), indices=unique(Signal$index)) {
  require(ggplot2)
  
  
  Signal <- subset(Signal, cab%in%cabs & index%in%indices)
    
  p <- ggplot(Signal, aes(x=index, y=signal)) + facet_wrap(~cab) + geom_bar(aes(group=lysate, fill=lysate), stat="identity") + geom_errorbar(aes(ymax=signal+error, ymin=signal-error), width=0.5)
  
  return(p)
  
}

#' \code{plotVariance} plots the signal vs. the standard error together
#' with a regression line or the error model, if available in Signal.
#' 
#' @param Signal a data.frame with columns "index" (numeric, the block numbering), "cab" (factor, indicating the capture antibody names), "signal" (numeric, the signal values), "lysate" (numeric, the fraction of lysate on the block), "error" (numeric, the standard error of the signal). If the data.frame contains "sigma", it is assumed that sigma represents a better estimate of the standard error based on an error model (see \code{getErrorModel})
#' @param cab character, the names of the capture antibodies to be plotted,
#' all if nothing is specified
#' @param index numeric, an index vector of the blocks to be plotted.
#' Defaults to all blocks.
#' @return a ggplot2 object
#' @export
plotVariance <- function(Signal,  cabs = unique(Signal$cab), indices=unique(Signal$index)) {
  require(ggplot2)
  
  Signal <- subset(Signal, cab%in%cabs & index%in%indices)
  
  p <- ggplot(Signal, aes(x=signal, y=error)) + facet_wrap(~cab) + geom_point(aes(group=lysate, color=lysate))
  
  if("sigma" %in% colnames(Signal)) {
    
    sortedSignal <- Signal[order(Signal$signal),]
    
    p <- p +
      geom_line(aes(x=signal, y=sigma, group=lysate, color=lysate), size=0.6, data=sortedSignal) +
      #stat_smooth(aes(x=signal, y=sigma), method="lm", formula=y~log(x)) +
      scale_y_log10()
    
    
  } else{
    p <- p + stat_smooth(method="lm", formula=y~log(x)) + scale_y_log10()
  }
  
  return(p)
  
  
}



#' \code{gerErrorModel} estimates error model parameters from a max-likelihood fit
#' 
#' @param Signal a data.frame with columns "index" (numeric, the block numbering), "cab" (factor, indicating the capture antibody names), "signal" (numeric, the signal values), "lysate" (numeric, the fraction of lysate on the block), "error" (numeric, the standard error of the signal)
#' @param factors character vector from which factor variable is constructed,
#' defaults to c("cab", "lysate"), i.e. one set of error parameters
#' for each combiination of cab and lysate.
#' @param errormodel character with the error model to be fitted.
#' Should contain "x", indicating the signal. The result of the formula should
#' be the variance.
#' Defaults to "s0^2 + srel^2*x^2".
#' @param pars named vector with the initial guesses for the error model parameters
#' @param df numeric, indicating the number of replicates that have been used
#' to compute the standard error.
#' @param ... futher arguments passed to \code{optim}.
#' @return data.frame with the original \code{Signal}, columns for the estimated error parameters and the final prediction of uncertainty, i.e. a \code{sigma} column of the signal uncertainty based on the error model parameters.
#' @export
getErrorModel <- function(Signal, factors=c("cab", "lysate"), errormodel="s0^2+srel^2*x^2", pars=c(s0=10, srel=0.1), df=100, ...) {
  
  conditionsColumn <- do.call(paste, as.data.frame(Signal[,factors]))
  conditions <- unique(conditionsColumn)
  Signal <- cbind(Signal, factor = conditionsColumn)
  
  out <- do.call(rbind, lapply(conditions, function(cond) {
    
    subdata <- subset(Signal, factor == cond)
    y <- subdata$error
    x <- subdata$signal
    
        
    nll <- function(p) {
      value <- with(as.list(p), {
        
        z <- eval(parse(text=errormodel))
        
        sum(log(z)-log(dchisq((df-1)*(y^2)/z, df=df-1)))

        
      })
      
      return(value)
      
    }
    
    fit <- optim(pars, nll, ...)
    pars <- abs(fit$par)

    
    data.frame(factor = cond, as.list(pars))
    
    
  }))
  errorPars <- as.data.frame(out[match(Signal$factor, out$factor),-1])
  colnames(errorPars) <- names(pars)
  

  sigma <- sqrt(with(cbind(errorPars, x=Signal$signal), eval(parse(text=errormodel))))

  
  
    
  
  
  Signal <- cbind(Signal, errorPars, sigma=sigma)
                 
  
  return(Signal)
  
    
  
}


#' \code{estimateParameters} estimates the calibration parameters and concentrantions
#' 
#' @param Signal a data.frame with columns "index" (numeric, the block numbering), "cab" (factor, indicating the capture antibody names), "signal" (numeric, the signal values), "lysate" (numeric, the fraction of lysate on the block), "condition" (numeric, the numbered lysate conditions), "sigma" (numeric, the standard error of the signal calculated by the error model. Alternatively, the user can specify other values, e.g. standard error of the mean.)
#' @param Calibrators data.frame with columns "index" (numeric, giving the block numbering), "purpose" (factor, only a subset of Calibrators with purpose=="target" is used. If Calibrators is the output of generateSetup, purpose will be either "calibrator" or "target".), "name" (factor, the target names), "concentration" (numeric, the known target concentrations due to the calibrators), "lysate" (numeric, the fraction of lysate on the block), "condition" (numeric, numbering the different lysate conditions. Indices without lysate must have condition 0!)
#' @param fitStructure data.frame with fixed values for the calibration parameters with the columns "cab" (factor, naming the capture antibodies), one numeric column for each target name (with the target name as column name), "saturation" (numeric, the saturation level, i.e. 1 for no saturation and >1 for saturation), "offset" (numeric, the signal background parameters).
#' @param verbose bool, print extra information in the console. Defaults to FALSE.
#' @return data.frame with columns "condition" (numeric, the lysate conditions), "purpose" (factor, usually "target"), "name" (factor, the target names), "concentration" (numeric, the estimated target concentrations), "sigma" (numeric, the estimated gaussian uncertainty of the concentrations).
#' @export
estimateParameters <- function(Signals, Calibrators, fitStructure, verbose=FALSE) {
  require(minpack.lm)
  require(corpcor)
  
  # Ensure correct order
  Signals <- Signals[order(Signals$cab, Signals$index),]
  Calibrators <- Calibrators[order(Calibrators$purpose, Calibrators$name, Calibrators$index),]
  
  # Generate matrices
  cabs <- unique(Signals$cab)
  targets <- unique(subset(Calibrators, purpose=="target")$name)
  conditions <- unique(Signals$condition)
  
  signalData <- as.matrix(do.call(cbind, lapply(cabs, function(t) subset(Signals, cab==t)$signal)))
  colnames(signalData) <- cabs

  
  sigmaData <- as.matrix(do.call(cbind, lapply(cabs, function(t) subset(Signals, cab==t)$sigma)))
  colnames(sigmaData) <- cabs
  
  caliData <- as.matrix(do.call(cbind, lapply(targets, function(t) subset(Calibrators, purpose=="target" & name==t)$concentration)))
  colnames(caliData) <- targets
  
  
  lysateData <- as.matrix(do.call(cbind, lapply(cabs, function(t) subset(Signals, cab==t)$lysate)))
  colnames(lysateData) <- cabs
  
  conditionData <- as.matrix(do.call(cbind, lapply(cabs, function(t) subset(Signals, cab==t)$condition)))
  colnames(conditionData) <- cabs
  
  if(verbose) {
    
    print(lysateData)
    print(caliData)
    
  }
  
    
  # weighted residuals function
  res <- function(M, Ai, deltai, concTarg) {
    
    #print(concTarg)
    
    targets <- colnames(M)
    totalConc <- concTarg*lysateData[,1]+caliData #matrices
    AiLysate <- 1/(t(1 - lysateData) + t(lysateData)*(1/Ai))
    signal <- t((M%*%t(totalConc) + deltai)*AiLysate)
    
    wr <- as.vector((signal - signalData)/sigmaData)
                
    return(wr)
    
  }
  
  
  # create translation between parameter names and matrices for res
  G <- expand.grid(cabs, targets)
  MNames <- paste("M", G[[1]], G[[2]], sep=".")
  
  deltaiNames <- paste("deltai", cabs, sep=".")
  AiNames <- paste("Ai", cabs, sep=".")
  allBlocks <- 1:dim(lysateData)[1]
                         
  G <- expand.grid(conditions, targets)
  concTargNames <- paste("conc", G[[1]], G[[2]], sep=".")
  
  
  

  # fixed parameters
  MFixed <- as.matrix(fitStructure[,!(colnames(fitStructure)%in%c("cab", "saturation", "offset"))])
  deltaiFixed <- fitStructure$offset
  AiFixed <- 1/fitStructure$saturation
  
  G <- expand.grid(fitStructure$cab, colnames(MFixed))
  MNamesFixed <- paste("M", G[[1]], G[[2]], sep=".")
  MFixed <- as.vector(MFixed); names(MFixed) <- MNamesFixed
  deltaiNamesFixed <- paste("deltai", fitStructure$cab, sep="."); names(deltaiFixed) <- deltaiNamesFixed
  AiNamesFixed <- paste("Ai", fitStructure$cab, sep="."); names(AiFixed) <- AiNamesFixed
  
  MFixed <- MFixed[!is.na(MFixed)]
  deltaiFixed <- deltaiFixed[!is.na(deltaiFixed)]
  AiFixed <- AiFixed[!is.na(AiFixed)]
  
  concFixed <- rep(0, length(targets))
  names(concFixed) <- paste("conc.0.", targets, sep="")
  
    
  pFixed <- c(MFixed, deltaiFixed, AiFixed, concFixed) 
  print(pFixed)
  

  # weighted residuals adapted to nls.lm
  resNls <- function(p) {
    
    p <- c(p, pFixed)
        
    M <- matrix(p[MNames], nrow=length(cabs), ncol=length(targets))
    colnames(M) <- targets
    rownames(M) <- cabs
    
    Ai <- p[AiNames]
    names(Ai) <- cabs
    
    deltai <- p[deltaiNames]
    names(deltai) <- cabs
    
    concTargPerCondition <- matrix(p[concTargNames], nrow=length(conditions), ncol=length(targets))
    
    
    
    concTarg <- concTargPerCondition[match(conditionData[,1], conditions),]
    
    colnames(concTarg) <- targets
    rownames(concTarg) <- allBlocks
    
    colnames(concTargPerCondition) <- targets
    rownames(concTargPerCondition) <- conditions
    
    
    
    out <- res(M, Ai, deltai, concTarg)
    attr(out, "M") <- M
    attr(out, "Ai") <- Ai
    attr(out, "deltai") <- deltai
    attr(out, "concTarg") <- concTargPerCondition
    
    
    return(out)
    
  }
  

  
  
  # nonlinear least squares estimation
  MNamesFree <- MNames[!(MNames%in%names(pFixed))]
  deltaiNamesFree <- deltaiNames[!(deltaiNames%in%names(pFixed))]
  AiNamesFree <- AiNames[!(AiNames%in%names(pFixed))]
  concNamesFree <- concTargNames[!(concTargNames%in%names(pFixed))]
  
  
  pnames <- c(MNamesFree,
              deltaiNamesFree,
              AiNamesFree,
              concNamesFree)
    
  Mini <- rep(1, length(MNamesFree))
  deltaiini <- rep(1, length(deltaiNamesFree))
  Aiini <- rep(1, length(AiNamesFree))
  concTargini <- rep(0, length(concNamesFree))
  
  ini <- c(Mini, deltaiini, Aiini, concTargini)
  names(ini) <- pnames
  
  
    
  cat("fitting ...")
  
  lower <- rep(-Inf, length(pnames))
  names(lower) <- pnames
  lower[concTargNames] <- 0
  lower[AiNames] <- 0
  
  fit <- nls.lm(fn=resNls, par=ini, control=list(maxiter=1e3, ftol=0))#, lower=rep(0, length(ini)))
  

  cat("finished\n")
 
  # recover parameters
  
    
  out <- resNls(fit$par)
  
  
    
  # get gaussian error estimation
  
  
  var <- diag(pseudoinverse(0.5*fit$hessian))
  names(var) <- colnames(fit$hessian)
  #var2 <- diag(vcov(fit))
  #print((var-var2)/var)
  
  
  if(verbose) {
    cat(paste(names(fit$par), ": ", fit$par, " ± ", sqrt(var), "\n", sep=""))
    cat("\n")
  }
  
  var <- var[names(var)%in%concNamesFree]
  sigma <- sqrt(matrix(var, nrow=length(conditions)-1, ncol=length(targets)))
  sigma <- rbind(0, sigma) 
  colnames(sigma) <- targets
    
  
  result <- attr(out, "concTarg")
    
  resultLong <- do.call(rbind, lapply(colnames(result), function(r) {
    data.frame(condition=conditions, purpose="target", name=r, concentration=result[,r], sigma = sigma[,r])
  }))
  
  attr(resultLong, "residuals") <- out*as.vector(sigmaData)
  attr(resultLong, "M") <- attr(out, "M")
  attr(resultLong, "deltai") <- attr(out, "deltai")
  attr(resultLong, "Ai") <- attr(out, "Ai")
    
  return(resultLong)
  
    
  
  
  
  
  
  
  
}
