## Functions

#' \code{generateSetup} uses a setup data.frame to generate
#' a latin hypercube sample of calibrator concentrations.
#' 
#' @param setup a data.frame with the entries target (character/factor),
#' wells (numeric), min (numeric), max (numeric) and a column for each
#' calibrator with its content of target protein.
#' @return a data.frame with the target samples (purpose = target) and the corresponding calibrator sample (purpose = calibrator).
#' Returns attributes "setup" with the setup dataframe,
#' "targetSample", a matrix of target concentrations and
#' "calibratorSample", a matrix of calibrator concentrations
#' @export
generateSetup <- function(setup) {
  
  require(lhs)
  
  M <- as.matrix(setup[,-(1:4)])
  M <- replace(M, is.na(M), 0)
  rownames(M) <- setup$target
 # print(M)
    
  wells <- max(as.numeric(setup$wells))
  ntargets <- length(unique(setup$target))
  
  blocks <- analyzeBlocks(M)
  
  fsample <- NULL
  fsample <- do.call(cbind, lapply(1:length(blocks), function(i) {
    
    sM <- M[blocks[[i]],]
    zeros <- apply(sM, 2, function(v) all(v==0))
    sM <- sM[,!zeros]
    
   # print(sM)
    
    vol <- det(sM)
    mysample <- improvedLHS(round(2*wells/vol), dim(sM)[1])
    colnames(mysample) <- rownames(sM)
    mysample <- do.call(cbind, lapply(colnames(mysample), function(t) {
      
      tmin <- setup[setup$target==t,"min"]
      tmax <- setup[setup$target==t,"max"]
      values <- mysample[,t]*(tmax-tmin) + tmin
      return(values)
      
    }))
    
    
    originals <- mysample%*%t(solve(sM))
    ok <- which(apply(originals, 1, function(v) all(v >= 0))==TRUE)
    mysample <- mysample[sample(ok, wells),]
    #print(mysample)
    
    colnames(mysample) <- rownames(sM)
    
    return(mysample)
    
  }))
  
  csample <- t(solve(M)%*%t(fsample)[rownames(M),])
  
  
  # convert to long format
  targets <- colnames(fsample)
  psampleLong <- do.call(rbind, lapply(targets, function(t) {
    data.frame(index=1:wells, purpose="target", name=t, concentration=fsample[,t])
  }))
  calibrators <- colnames(csample)
  csampleLong <- do.call(rbind, lapply(calibrators, function(t) {
    data.frame(index=1:wells, purpose="calibrator", name=t, concentration=csample[,t])
  }))
  

  
  
  
  result <- rbind(psampleLong, csampleLong)
  
  attr(result, "setup") <- setup
  attr(result, "targetSample") <- fsample
  attr(result, "calibratorSample") <- csample
            
  
  
  return(result)
     
  
}

## Analyze matrix with zero and one entries for connection components
analyzeBlocks <- function(M) {
  out <- which(apply(M, 1, sum)==0)
  if(length(out) > 0) {
    M <- M[-out,]
    cat("matrix contains zero rows which have been eliminated\n")
  }
  n <- dim(M)[1]
  rcomponents <- list()
  ccomponents <- list()
  
  counter <- 0
  while(length(unlist(rcomponents))< n) {
    
    counter <- counter + 1
    
    if(length(unlist(rcomponents)) == 0) {
      w <- 1
    } else {
      mysample <- (1:n)[-unlist(rcomponents)]
      w <- min(mysample)
    }
    
    repeat {  
      
      v <- unique(rapply(as.list(w), function(i) which(M[i,] != 0)))
      wnew <- unique(rapply(as.list(v), function(j) which(M[,j] != 0)))
      if(length(wnew) == length(w)) break
      w <- wnew
    }
    rcomponents[[counter]] <- w
    ccomponents[[counter]] <- v
    
    
  }
  
  #print(M[unlist(rcomponents), unlist(ccomponents)])
  
  
  return(rcomponents)
  
  
  
  
}



#' \code{plotSetup} Takes the output of generateSetup and plots it
#' 
#' @param out result from generateSetup
#' @param what either "target" or "calibrator")
#' indicating the entry to be plotted
#' @return a ggplot2 object
#' @export
plotSetup <- function(out, what="target") {
  
  require(ggplot2)
  out <- subset(out, purpose==what)
  targets <- unique(out$name)
  combs <- combn(targets, 2, simplify=FALSE)
  data <- do.call(rbind, lapply(combs, function(t) {
    x1 <- subset(out, name==t[1])$concentration
    x2 <- subset(out, name==t[2])$concentration
    d <- data.frame(what = paste(t, collapse=" - "), x1 = x1, x2 = x2)
    return(d)
  }))
  
  g <- ggplot(data, aes(x=x1, y=x2)) +
    facet_wrap(~ what, scales="free") + xlab("concentration") + ylab("concentration") +
    geom_point()
  
  return(g)
  
}