#' \code{read.gpr} reads a gpr-file and returns a data.frame
#' 
#' @param file character, the file name
#' @param header bool, set to TRUE if the file contains a header, defaults to TRUE
#' @param sep character, the separation character used within the text file to separate columns. Defaults to tab.
#' @param ... other arguments passed to the \code{read.table} function
#' @return a data.frame
#' @details Usually, the gpr file is a text file with some lines of additional information before the header and table information come. The first lines are skipped.
#' @export
read.gpr <- function(file, header=TRUE, sep="\t", ...) {
  
  # check how many lines need to be skipped at the beginning of the gpr file
  counts <- count.fields(file)
  myskip <- which(counts==median(counts))[1] - 1
  
  data <- read.table(file, skip=myskip, header=header, sep=sep, ...)
  
  return(data)
  
  
}

#' \code{getSignals} extracts the signals per block for each target
#' 
#' @param signalData data.frame, usually produced by \code{read.gpr}, with columns "Block" (the block numbers), "Column" (the slide columns), "Row" (the slide rows), "ID" (the target ID, usually the capture-antibody names) and columns for the signal foreground and local background.
#' @param normalizerData data.frame (see signalData). Defaults to \code{NULL}. In this case, the signal is not normalized.
#' @param fg character, the column name of the signal foreground to be used. Defaults to "F700.Median".
#' @param bg character, the column name of the local signal background. Defaults to "B700.Median".
#' @param norm character, the column name of the foreground signal in the normalizer data.frame used for normalization purposes. Defaults to "F800.Median".
#' @param badFlags numeric, the identifier of spots to be omitted in the signal computation. Defaults to -100.
#' 
#' @return data.frame with columns "index" (numeric, the block number), "cab" (factor, the capture antibody names), "signal" (numeric, the signal values), "lysate" (numeric, 0, needs to be set by the slide description, see \code{makeArray}), "condition" (numeric, 0, needs to be set by the slide description, see \code{makeArray}), "error" (numeric, the standard error of the signal values)
#' @details Signals from different spots (encoded in the \code{Row} and \code{Column} columns of \code{signalData}) are averaged by computing the arithmetic mean. The error of the signal is the standard error, i.e. the standard deviation divided by the square root of the number of replica.
#' @export
getSignals <- function(signalData, normalizerData=NULL, fg="F700.Median", bg="B700.Median", norm="F800.Median", badFlags = -100) {
  
  # correct for local background and normalization
  if(is.null(normalizerData)) {
    
    signal <- data.frame(Block = signalData$Block,
                         Column = signalData$Column,
                         Row = signalData$Row,
                         ID = signalData$ID,
                         Signal = signalData[,fg]-signalData[,bg])

    
  } else {
    
    signal <- data.frame(Block = signalData$Block,
                         Column = signalData$Column,
                         Row = signalData$Row,
                         ID = signalData$ID,
                         Signal = mean(normalizerData[,norm], na.rm=TRUE)*(signalData[,fg]-signalData[,bg])/normalizerData[,norm])
    
  }
  
  signal[signalData$Flags%in%badFlags, "Signal"] <- NA
  
  
  
  
  # Reduce signal set by averaging
  
  blocks <- unique(signal$Block)
  ids <- unique(signal$ID)
  
  redSignal <- NULL
  for(b in blocks) {
    for(i in ids) {
      
      subSignal <- subset(signal, ID == i & Block == b)
      meanSignal <- mean(subSignal$Signal, na.rm=TRUE)
      nSignal <- length(which(!is.na(subSignal$Signal)))
      sdSignal <- sd(subSignal$Signal, na.rm=TRUE)
      errorSignal <- sdSignal/sqrt(nSignal)
      
      out <- data.frame(index=b,
                        cab = i,
                        signal = meanSignal,
                        lysate=0,
                        condition=0,
                        error = errorSignal)
      
      redSignal <- rbind(redSignal, out)
      
      
    }
  }
  
  return(redSignal)
  
  
}

makeArray <- function(slideDescription, rawSignals) {
  
  ## Translate the slide description into the Calibration format
 
  calibrators1 <- data.frame(index = slideDescription$Block,
                             purpose = "target",
                             name = slideDescription$AB2,
                             concentration = slideDescription$Concentration,
                             lysate = 0,
                             time = slideDescription$Time,
                             substance = slideDescription$Substance)
  
  
  
  conditions <- unique(paste(slideDescription$Time, slideDescription$Substance))
  names(conditions) <- 1:length(conditions)-1
  
  ind <- match(paste(calibrators1$time, calibrators1$substance), conditions)
  
  calibrators1 <- cbind(calibrators1, condition = as.numeric(names(conditions)[ind]))
  
  calibratorBlocks <- unique(slideDescription$Block[slideDescription$Purpose=="calibrator"])
  lysateBlocks <- unique(slideDescription$Block)[!unique(slideDescription$Block)%in%calibratorBlocks]
  
  calibrators1$lysate[calibrators1$index%in%lysateBlocks] <- 1
  
  if("lysate" %in% colnames(slideDescription)) calibrators1$lysate <- slideDescription$lysate
  
  
  ## Translate the rawSignals into Signals format
  
  signal1 <- rawSignals
  signal1$condition <- calibrators1$condition[match(signal1$index, calibrators1$index)]
  signal1$lysate <- calibrators1$lysate[match(signal1$index, calibrators1$index)]
  
  return(list(Signals=signal1, Calibrators=calibrators1))
  
}


plotFit <- function(myfit, Calibrators) {
  
  translation <- NULL
  for(i in 1:dim(myfit)[1]) {
    
    myrow <- which(Calibrators$condition==myfit[i,"condition"])[1]
    translation <- rbind(translation, Calibrators[myrow, c("time", "substance")])
    
    
  }
  colnames(translation) <- c("time", "substance")
  
  myfit <- cbind(myfit, translation)
  
  ggplot(myfit, aes(x=time, y=concentration, group=substance, color=substance, fill=substance)) + facet_wrap(~name, scales="free") + geom_point() + 
    #geom_errorbar(aes(ymax=concentration+sigma, ymin=concentration-sigma), width=0) + geom_line() + 
    geom_ribbon(aes(ymax=concentration+sigma, ymin=concentration-sigma), alpha=0.3) +
    expand_limits(y=0)
  
}
